(*Based on Bertram Felgenhauer's Haskell code of
haskell-rewriting/term-rewriting*)
module Fun = Function;;
module Var = Variable;;
module UF = Util.UnionFind;;
module VMap = Util.Map.Make (Var) (Util.Int);;

open Prelude;;
open Util;;

(*** MODULES ******************************************************************)
module Make (L : LABEL) = struct

module Subst = Substitution.Make (L);;
module Term = Term.Make (L);;
 
type term = Term.t = Var of Var.t | Fun of Fun.t * term list;;

module Annot = struct
  type t = VarA of Var.t | FunA of Fun.t * int list | FunP of Fun.t * Term.t list;;
  let default = VarA Var.zero;;
end
include Annot;;

module UM = UnionFindMonad.Make (Annot);;

module UnifyM = Util.Monad.Transformer.State (VMap) (UM);;

let funari = function
| FunA (f, ns) -> (f, List.length ns)
| FunP (f, ts) -> (f, List.length ts)
| _ -> failwith "funari is not defined on variable annotations"
;;

let (>>=) = UnifyM.(>>=);;
let (>>) = UnifyM.(>>);;
let return = UnifyM.return;;
let new_node = UnifyM.lift <.> UM.new_node;;
let lookup = UnifyM.lift <.> UM.lookup;;
let merge f i j = UnifyM.lift (UM.merge f i j);;
let annotate n = UnifyM.lift <.> UM.annotate n;;

(*Create a new node representing a given term. Variable nodes are shared
whenever possible. Function applications will be pending initially.*)
let mk_node = function
| Term.Var x ->
  UnifyM.get >>= fun m ->
  if VMap.mem x m then
    return (VMap.find x m)
  else
    new_node (VarA x) >>= fun n ->
    UnifyM.update (VMap.add x n) >>
    return n
| Term.Fun (f, ts) -> new_node (FunP (f, ts))
;;

let expand n = function
| FunP (f, ts) ->
  UnifyM.map mk_node ts >>= fun ns ->
  let l = FunA (f, ns) in
  annotate n l >>
  return l
| l -> return l
;;

let rec solve = function
| [] -> return true
| (t, u) :: xs ->
  lookup t >>= fun (t, t') ->
  lookup u >>= fun (u, u') ->
  if t = u then solve xs else
  match (t', u') with
  | (VarA _, s) | (s, VarA _) ->
    merge (fun _ _ -> s) t u >>
    solve xs
  | _ when funari t' = funari u' ->
    expand t t' >>= function (FunA (_, ts)) ->
    expand u u' >>= function (FunA (_, us)) ->
    merge (const) t u >>
    solve (List.rev_append (List.rev_zip ts us) xs)
  | _ -> return false
;;

let acyclic size succs root =
  let color = Array.make size 0 in
  let rec dfs n =
    let c = color.(n) in
    if c = 0 then begin
      color.(n) <- 1;
      if List.for_all dfs (succs n) then begin
        color.(n) <- 2;
        true
      end else
        false
    end else if c = 1 then false
    else true (* all successors processed *)
  in
  dfs root
;;

let unify t u =
  let act =
    mk_node t >>= fun t' ->
    mk_node u >>= fun u' ->
    solve [(t', u')] >>= fun success ->
    return (t', success)
  in
  let (union, ((root, success), vmap)) = UM.run' (UnifyM.eval VMap.empty act) in
  let succs n = match snd (UF.lookup union n) with
    | VarA x -> []
    | FunA (f, ns) -> ns
    | FunP (f, ts) ->
      List.foldl (fun ns t -> List.foldl (fun ns x ->
        if VMap.mem x vmap then VMap.find x vmap :: ns
        else ns) ns (Term.vars t)) [] ts
  in
  if success && acyclic (UF.size union) succs root then
    let cache = Hashtbl.create 16 in
    let rec apply_vmap = function
      | Term.Var x as v ->
        if VMap.mem x vmap then lookup_node (VMap.find x vmap) else v
      | Term.Fun (f, ts) -> Term.Fun (f, List.map apply_vmap ts)
    and mk_term = function
      | VarA x -> Term.Var x
      | FunA (f, ns) -> Term.Fun (f, List.map (lookup_node) ns)
      | FunP (f, ts) -> apply_vmap (Term.Fun (f, ts))
    and lookup_node n =
      if Hashtbl.mem cache n then Hashtbl.find cache n
      else
        let t = mk_term (UF.label union (fst (UF.lookup union n))) in
        (Hashtbl.add cache n t; t)
    in
    let subst =
      Subst.of_list (List.map (fun (x, n) -> (x, lookup_node n)) (VMap.to_list vmap)) in
    Some subst
  else
    None
;;

end
