(*Based on Bertram Felgenhauer's Haskell code of
haskell-rewriting/union-find-array*)
open Prelude;;

type node = int;;

(* A collection of disjoint sets of integers (represented by trees, s.t., the
root of a tree is the canonical representative of the equivalence class that
constitutes of its nodes.

up.(i)    points to the parent of node i
rank.(i)  is the depth of the tree with root i
label.(i) is the label of node i
def       is the default value used for labels *)
type 'a t = {
  up : node array;
  rank : int array;
  label : 'a array;
  def : 'a;
}

let mk s d = {
  up = Array.init s id;
  rank = Array.make s 0;
  label = Array.make s d;
  def = d;
}

let size u = Array.length u.up;;

let label u n = u.label.(n);;

let annotate u i v = u.label.(i) <- v;;

(*lookup with built-in path compression*)
let rec lookup' u i =
  let i' = u.up.(i) in
  if i = i' then  i
  else begin
    let i'' = lookup' u i' in
    u.up.(i) <- i'';
    i''
  end
;;

let lookup u i =
  let i' = lookup' u i in
  let l' = u.label.(i') in
  (i', l')
;;

let merge u f a b =
  let (a', va) = lookup u a in
  let (b', vb) = lookup u b in
  if a' = b' then None else begin
    let ra = u.rank.(a') in
    let rb = u.rank.(b') in
    let cont x vx y vy =
      (*the label at y should no longer be accessible*)
      let (v, w) = f vx vy in
      u.label.(x) <- v;
      Some w
    in
    if ra < rb then begin
      u.up.(a') <- b';
      cont b' vb a' va
    end else if ra > rb then begin
      u.up.(b') <- a';
      cont a' va b' vb
    end else begin
      u.up.(a') <- b';
      u.rank.(b') <- ra + 1;
      cont b' vb a' va
    end
  end
;;

let grow' u size' =
  let up' = Array.init size' (fun x -> x) in
  let rank' = Array.make size' 0 in
  let label' = Array.make size' u.def in
  for i = 0 to size u - 1 do
    up'.(i) <- u.up.(i);
    rank'.(i) <- u.rank.(i);
    label'.(i) <- u.label.(i)
  done;
  { u with up = up'; rank = rank'; label = label' }
;;

let grow u s = if s <= size u then u else grow' u s;;

(*
let flatten u =
  for i = 0 to size u - 1 do
    ignore (lookup' u i)
  done
;;

let copy u = grow' u (size u);;
*)
